/* glplib07.c (stream input/output) */

/***********************************************************************
*  This code is part of GLPK (GNU Linear Programming Kit).
*
*  Copyright (C) 2000, 01, 02, 03, 04, 05, 06, 07, 08 Andrew Makhorin,
*  Department for Applied Informatics, Moscow Aviation Institute,
*  Moscow, Russia. All rights reserved. E-mail: <mao@mai2.rcnet.ru>.
*
*  GLPK is free software: you can redistribute it and/or modify it
*  under the terms of the GNU General Public License as published by
*  the Free Software Foundation, either version 3 of the License, or
*  (at your option) any later version.
*
*  GLPK is distributed in the hope that it will be useful, but WITHOUT
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
*  License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with GLPK. If not, see <http://www.gnu.org/licenses/>.
***********************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _GLPSTD_STDIO
#include "glplib.h"

#if 0
/***********************************************************************
*  NAME
*
*  xfopen - open file
*
*  SYNOPSIS
*
*  #include "glplib.h"
*  FILE *xfopen(const char *fname, const char *mode);
*
*  DESCRIPTION
*
*  The routine xfopen opens a file using the character string fname as
*  the file name and the character string mode as the open mode.
*
*  RETURNS
*
*  If the file has been open successfully, the routine xfopen returns a
*  pointer to an i/o stream associated with the file (i.e. a pointer to
*  an object of the FILE type). Otherwise the routine return NULL. */

FILE *xfopen(const char *fname, const char *mode)
{     LIBENV *env = lib_link_env();
      int k;
      /* find free slot */
      for (k = 0; k < LIB_MAX_OPEN; k++)
         if (env->file_slot[k] == NULL) break;
      if (k == LIB_MAX_OPEN)
         xfault("xfopen: too many open files\n");
      /* open a file and store a pointer to the i/o stream */
      env->file_slot[k] = fopen(fname, mode);
      return env->file_slot[k];
}

/***********************************************************************
*  NAME
*
*  xfclose - close file
*
*  SYNOPSIS
*
*  #include "glplib.h"
*  void xfclose(FILE *fp);
*
*  DESCRIPTION
*
*  The routine xfclose closes a file associated with i/o stream, which
*  the parameter fp points to. It is assumed that the file was open by
*  the routine xfopen. */

void xfclose(FILE *fp)
{     LIBENV *env = lib_link_env();
      int k;
      /* check if the i/o stream pointer is valid */
      if (fp == NULL)
         xfault("xfclose: fp = %p; null i/o stream\n", fp);
      for (k = 0; k < LIB_MAX_OPEN; k++)
         if (env->file_slot[k] == fp) break;
      if (k == LIB_MAX_OPEN)
         xfault("xfclose: fp = %p; invalid i/o stream\n", fp);
      /* close a file and free the corresponding slot */
      fclose(fp);
      env->file_slot[k] = NULL;
      return;
}
#endif

/**********************************************************************/

XFILE *xfopen(const char *fname, const char *mode)
{     LIBENV *env = lib_link_env();
      XFILE *file = NULL;
      FILE *fp;
      fp = fopen(fname, mode);
      if (fp == NULL) goto done;
      file = xmalloc(sizeof(XFILE));
      file->fname = xmalloc(strlen(fname)+1);
      strcpy(file->fname, fname);
      file->fp = fp;
      file->prev = NULL;
      file->next = env->file_ptr;
      if (file->next != NULL) file->next->prev = file;
      env->file_ptr = file;
done: return file;
}

/**********************************************************************/

int xferror(XFILE *file)
{     FILE *fp = file->fp;
      return ferror(fp);
}

/**********************************************************************/

int xfeof(XFILE *file)
{     FILE *fp = file->fp;
      return feof(fp);
}

/**********************************************************************/

int xfgetc(XFILE *file)
{     FILE *fp = file->fp;
      return fgetc(fp);
}

/**********************************************************************/

int xfputc(int c, XFILE *file)
{     FILE *fp = file->fp;
      return fputc(c, fp);
}

/***********************************************************************
*  NAME
*
*  xdoprnt - perform formatted output (basic routine)
*
*  SYNOPSIS
*
*  #include "glplib.h"
*  int xdoprnt(int (*func)(void *info, int c), void *info, const char
*     *fmt, va_list arg);
*
*  DESCRIPTION
*
*  The routine xdoprnt is a basic routine to perform formatted output.
*  It is equivalent to the standard routine vfprintf, except that the
*  output is passed to the formal routine func rather than to an output
*  stream.
*
*  The formal routine func is called every time a next character should
*  be written. If there is an output error, it should return a negative
*  value, in which case the output is terminated.
*
*  RETURNS
*
*  The routine xdoprnt returns the number of characters passed to the
*  formal routine func. However, if there was an output error (detected
*  by the formal routine func), a negative value is returned. */

int xdoprnt(int (*func)(void *info, int c), void *info, const char
      *fmt, va_list arg)
{     int cnt, j;
      char buf[4000+1];
#ifdef HAVE_VSNPRINTF
      cnt = vsnprintf(buf, sizeof(buf), fmt, arg);
#else
      cnt = vsprintf(buf, fmt, arg);
#endif
      xassert(0 <= cnt && cnt < sizeof(buf));
      xassert((int)strlen(buf) == cnt);
      for (j = 0; j < cnt; j++)
      {  if (func(info, (unsigned char)buf[j]) < 0)
         {  cnt = -1;
            break;
         }
      }
      return cnt;
}

/**********************************************************************/

static int my_put(void *file, int c)
{     return
         xfputc(c, file);
}

int xfprintf(XFILE *file, const char *fmt, ...)
{     va_list arg;
      int ret;
      va_start(arg, fmt);
#if 0
      ret = vfprintf(file->fp, fmt, arg);
#else
      ret = xdoprnt(my_put, file, fmt, arg);
#endif
      va_end(arg);
      return ret;
}

/**********************************************************************/

int xfflush(XFILE *file)
{     FILE *fp = file->fp;
      return fflush(fp);
}

/**********************************************************************/

int xfclose(XFILE *file)
{     LIBENV *env = lib_link_env();
      FILE *fp = file->fp;
      if (file->prev == NULL)
         env->file_ptr = file->next;
      else
         file->prev->next = file->next;
      if (file->next == NULL)
         ;
      else
         file->next->prev = file->prev;
      xfree(file->fname);
      xfree(file);
      return fclose(fp);
}

/**********************************************************************/

int lib_open_log(const char *fname)
{     /* open hardcopy file */
      LIBENV *env = lib_link_env();
      if (env->log_file != NULL)
      {  /* hardcopy file is already open */
         return 1;
      }
      env->log_file = fopen(fname, "w");
      if (env->log_file == NULL)
      {  /* cannot create hardcopy file */
         return 2;
      }
      setvbuf(env->log_file, NULL, _IOLBF, BUFSIZ);
      return 0;
}

/**********************************************************************/

int lib_close_log(void)
{     /* close hardcopy file */
      LIBENV *env = lib_link_env();
      if (env->log_file == NULL)
      {  /* hardcopy file is already closed */
         return 1;
      }
      fclose(env->log_file);
      env->log_file = NULL;
      return 0;
}

/* eof */
